@@ -1,44 +1,36 @@
+// SPDX-License-Identifier: Apache-2.0 WITH LLVM-exception
 // -*- Mode: C++ -*-
 //
-// Copyright (C) 2013-2020 Red Hat, Inc.
-//
-// This file is part of the GNU Application Binary Interface Generic
-// Analysis and Instrumentation Library (libabigail).  This library is
-// free software; you can redistribute it and/or modify it under the
-// terms of the GNU Lesser General Public License as published by the
-// Free Software Foundation; either version 3, or (at your option) any
-// later version.
-
-// This library is distributed in the hope that it will be useful, but
-// WITHOUT ANY WARRANTY; without even the implied warranty of
-// MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
-// General Lesser Public License for more details.
-
-// You should have received a copy of the GNU Lesser General Public
-// License along with this program; see the file COPYING-LGPLV3.  If
-// not, see <http://www.gnu.org/licenses/>.
+// Copyright (C) 2013-2022 Red Hat, Inc.
 //
 // Author: Dodji Seketeli
 
 /// @file
 
+#include "config.h"
 #include <cstring>
-#include <vector>
-#include <string>
 #include <iostream>
+#include <memory>
+#include <string>
+#include <vector>
 #include "abg-config.h"
 #include "abg-comp-filter.h"
 #include "abg-suppression.h"
 #include "abg-tools-utils.h"
 #include "abg-reader.h"
 #include "abg-dwarf-reader.h"
+#ifdef WITH_CTF
+#include "abg-ctf-reader.h"
+#endif
 
 using std::vector;
 using std::string;
 using std::ostream;
 using std::cout;
 using std::cerr;
-using abg_compat::shared_ptr;
+using std::shared_ptr;
+using abg_compat::optional;
+using namespace abigail;
 using abigail::ir::environment;
 using abigail::ir::environment_sptr;
 using abigail::translation_unit;
@@ -54,7 +46,7 @@ using abigail::comparison::get_default_harmful_categories_bitmap;
 using abigail::suppr::suppression_sptr;
 using abigail::suppr::suppressions_type;
 using abigail::suppr::read_suppressions;
-using namespace abigail::dwarf_reader;
+
 using abigail::tools_utils::emit_prefix;
 using abigail::tools_utils::check_file;
 using abigail::tools_utils::guess_file_type;
@@ -63,6 +55,9 @@ using abigail::tools_utils::gen_suppr_spec_from_kernel_abi_whitelists;
 using abigail::tools_utils::load_default_system_suppressions;
 using abigail::tools_utils::load_default_user_suppressions;
 using abigail::tools_utils::abidiff_status;
+using abigail::tools_utils::create_best_elf_based_reader;
+
+using namespace abigail;
 
 struct options
 {
@@ -83,10 +78,12 @@ struct options
   vector<string>	headers_dirs2;
   vector<string>        header_files2;
   bool			drop_private_types;
+  optional<bool>	exported_interfaces_only;
   bool			linux_kernel_mode;
   bool			no_default_supprs;
   bool			no_arch;
   bool			no_corpus;
+  bool			ignore_soname;
   bool			leaf_changes_only;
   bool			fail_no_debug_info;
   bool			show_hexadecimal_values;
@@ -111,9 +108,20 @@ struct options
   bool			show_redundant_changes;
   bool			show_symbols_not_referenced_by_debug_info;
   bool			show_impacted_interfaces;
+  bool			assume_odr_for_cplusplus;
+  bool			leverage_dwarf_factorization;
   bool			dump_diff_tree;
   bool			show_stats;
   bool			do_log;
+#ifdef WITH_DEBUG_SELF_COMPARISON
+  bool			do_debug_self_comparison;
+#endif
+#ifdef WITH_DEBUG_TYPE_CANONICALIZATION
+  bool			do_debug_type_canonicalization;
+#endif
+#ifdef WITH_CTF
+  bool			use_ctf;
+#endif
   vector<char*> di_root_paths1;
   vector<char*> di_root_paths2;
   vector<char**> prepared_di_root_paths1;
@@ -128,6 +136,7 @@ struct options
       no_default_supprs(),
       no_arch(),
       no_corpus(),
+      ignore_soname(false),
       leaf_changes_only(),
       fail_no_debug_info(),
       show_hexadecimal_values(),
@@ -152,9 +161,23 @@ struct options
       show_redundant_changes(),
       show_symbols_not_referenced_by_debug_info(true),
       show_impacted_interfaces(),
+      assume_odr_for_cplusplus(true),
+      leverage_dwarf_factorization(true),
       dump_diff_tree(),
       show_stats(),
       do_log()
+#ifdef WITH_CTF
+    ,
+      use_ctf()
+#endif
+#ifdef WITH_DEBUG_SELF_COMPARISON
+    ,
+      do_debug_self_comparison()
+#endif
+#ifdef WITH_DEBUG_TYPE_CANONICALIZATION
+    ,
+      do_debug_type_canonicalization()
+#endif
   {}
 
   ~options()
@@ -190,6 +213,9 @@ display_usage(const string& prog_name, ostream& out)
     << " --header-file2|--hf2 <path>  the path to one header of file2\n"
     << " --drop-private-types  drop private types from "
     "internal representation\n"
+    << "  --exported-interfaces-only  analyze exported interfaces only\n"
+    << "  --allow-non-exported-interfaces  analyze interfaces that "
+    "might not be exported\n"
     << " --no-linux-kernel-mode  don't consider the input binaries as "
        "linux kernel binaries\n"
     << " --kmi-whitelist|-w  path to a "
@@ -200,6 +226,7 @@ display_usage(const string& prog_name, ostream& out)
        "default suppression specification\n"
     << " --no-architecture  do not take architecture in account\n"
     << " --no-corpus-path  do not take the path to the corpora into account\n"
+    << " --ignore-soname  do not take the SONAMEs into account\n"
     << " --fail-no-debug-info  bail out if no debug info was found\n"
     << " --leaf-changes-only|-l  only show leaf changes, "
     "so no change impact analysis (implies --redundant)\n"
@@ -236,9 +263,23 @@ display_usage(const string& prog_name, ostream& out)
     << " --no-redundant  do not display redundant changes "
     "(this is the default)\n"
     << " --impacted-interfaces  display interfaces impacted by leaf changes\n"
+    << " --no-leverage-dwarf-factorization  do not use DWZ optimisations to "
+    "speed-up the analysis of the binary\n"
+    << " --no-assume-odr-for-cplusplus  do not assume the ODR to speed-up the "
+    "analysis of the binary\n"
     << " --dump-diff-tree  emit a debug dump of the internal diff tree to "
     "the error output stream\n"
     <<  " --stats  show statistics about various internal stuff\n"
+#ifdef WITH_CTF
+    << " --ctf use CTF instead of DWARF in ELF files\n"
+#endif
+#ifdef WITH_DEBUG_SELF_COMPARISON
+    << " --debug-self-comparison debug the process of comparing "
+    "an ABI corpus against itself"
+#endif
+#ifdef WITH_DEBUG_TYPE_CANONICALIZATION
+    << " --debug-tc debug the type canonicalization process"
+#endif
     << " --verbose show verbose messages about internal stuff\n";
 }
 
@@ -389,12 +430,18 @@ parse_command_line(int argc, char* argv[], options& opts)
 	}
       else if (!strcmp(argv[i], "--drop-private-types"))
 	opts.drop_private_types = true;
+      else if (!strcmp(argv[i], "--exported-interfaces-only"))
+	opts.exported_interfaces_only = true;
+      else if (!strcmp(argv[i], "--allow-non-exported-interfaces"))
+	opts.exported_interfaces_only = false;
       else if (!strcmp(argv[i], "--no-default-suppression"))
 	opts.no_default_supprs = true;
       else if (!strcmp(argv[i], "--no-architecture"))
 	opts.no_arch = true;
       else if (!strcmp(argv[i], "--no-corpus-path"))
 	opts.no_corpus = true;
+      else if (!strcmp(argv[i], "--ignore-soname"))
+	opts.ignore_soname = true;
       else if (!strcmp(argv[i], "--fail-no-debug-info"))
 	opts.fail_no_debug_info = true;
       else if (!strcmp(argv[i], "--leaf-changes-only")
@@ -576,12 +623,28 @@ parse_command_line(int argc, char* argv[], options& opts)
 	opts.show_redundant_changes = false;
       else if (!strcmp(argv[i], "--impacted-interfaces"))
 	opts.show_impacted_interfaces = true;
+      else if (!strcmp(argv[i], "--no-leverage-dwarf-factorization"))
+	opts.leverage_dwarf_factorization = false;
+      else if (!strcmp(argv[i], "--no-assume-odr-for-cplusplus"))
+	opts.leverage_dwarf_factorization = false;
       else if (!strcmp(argv[i], "--dump-diff-tree"))
 	opts.dump_diff_tree = true;
       else if (!strcmp(argv[i], "--stats"))
 	opts.show_stats = true;
       else if (!strcmp(argv[i], "--verbose"))
 	opts.do_log = true;
+#ifdef WITH_CTF
+      else if (!strcmp(argv[i], "--ctf"))
+        opts.use_ctf = true;
+#endif
+#ifdef WITH_DEBUG_SELF_COMPARISON
+      else if (!strcmp(argv[i], "--debug-self-comparison"))
+	opts.do_debug_self_comparison = true;
+#endif
+#ifdef WITH_DEBUG_TYPE_CANONICALIZATION
+      else if (!strcmp(argv[i], "--debug-tc"))
+	opts.do_debug_type_canonicalization = true;
+#endif
       else
 	{
 	  if (strlen(argv[i]) >= 2 && argv[i][0] == '-' && argv[i][1] == '-')
@@ -682,6 +745,8 @@ set_diff_context_from_opts(diff_context_sptr ctxt,
   ctxt->show_added_vars(opts.show_all_vars || opts.show_added_vars);
   ctxt->show_linkage_names(opts.show_linkage_names);
   ctxt->show_locs(opts.show_locs);
+  // Intentional logic flip of ignore_soname
+  ctxt->show_soname_change(!opts.ignore_soname);
   // So when we are showing only leaf changes, we want to show
   // redundant changes because of this: Suppose several functions have
   // their return type changed from void* to int*.  We want them all
@@ -747,6 +812,23 @@ set_diff_context_from_opts(diff_context_sptr ctxt,
   ctxt->dump_diff_tree(opts.dump_diff_tree);
 }
 
+/// Set a bunch of tunable buttons on the ELF-based reader from the
+/// command-line options.
+///
+/// @param rdr the reader to tune.
+///
+/// @param opts the command line options.
+static void
+set_generic_options(abigail::elf_based_reader& rdr, options& opts)
+{
+  rdr.options().show_stats = opts.show_stats;
+  rdr.options().do_log = opts.do_log;
+  rdr.options().leverage_dwarf_factorization =
+    opts.leverage_dwarf_factorization;
+  rdr.options().assume_odr_for_cplusplus =
+    opts.assume_odr_for_cplusplus;
+}
+
 /// Set suppression specifications to the @p read_context used to load
 /// the ABI corpus from the ELF/DWARF file.
 ///
@@ -758,14 +840,13 @@ set_diff_context_from_opts(diff_context_sptr ctxt,
 /// @param read_ctxt the read context to apply the suppression
 /// specifications to.  Note that the type of this parameter is
 /// generic (class template) because in practise, it can be either an
-/// abigail::dwarf_reader::read_context type or an
-/// abigail::xml_reader::read_context type.
+/// dwarf::read_context type or an
+/// abigail::abiabixml_reader::reader type.
 ///
 /// @param opts the options where to get the suppression
 /// specifications from.
-template<class ReadContextType>
 static void
-set_suppressions(ReadContextType& read_ctxt, const options& opts)
+set_suppressions(abigail::fe_iface& reader, const options& opts)
 {
   suppressions_type supprs;
   for (vector<string>::const_iterator i = opts.suppression_paths.begin();
@@ -773,7 +854,7 @@ set_suppressions(ReadContextType& read_ctxt, const options& opts)
        ++i)
     read_suppressions(*i, supprs);
 
-  if (read_context_get_path(read_ctxt) == opts.file1
+  if (reader.corpus_path() == opts.file1
       && (!opts.headers_dirs1.empty() || !opts.header_files1.empty()))
     {
       // Generate suppression specification to avoid showing ABI
@@ -793,7 +874,7 @@ set_suppressions(ReadContextType& read_ctxt, const options& opts)
 	}
     }
 
-  if (read_context_get_path(read_ctxt) == opts.file2
+  if (reader.corpus_path() == opts.file2
       && (!opts.headers_dirs2.empty() || !opts.header_files2.empty()))
     {
       // Generate suppression specification to avoid showing ABI
@@ -819,7 +900,7 @@ set_suppressions(ReadContextType& read_ctxt, const options& opts)
 
   supprs.insert(supprs.end(), wl_suppr.begin(), wl_suppr.end());
 
-  add_read_context_suppressions(read_ctxt, supprs);
+  reader.add_suppressions(supprs);
 }
 
 /// Configure the abigail::xml_reacher::read_context based on the
@@ -829,11 +910,11 @@ set_suppressions(ReadContextType& read_ctxt, const options& opts)
 ///
 /// @param opts the command-line options to configure @p ctxt from.
 static void
-set_native_xml_reader_options(abigail::xml_reader::read_context& ctxt,
+set_native_xml_reader_options(abigail::fe_iface& rdr,
 			      const options& opts)
 {
-  consider_types_not_reachable_from_public_interfaces(ctxt,
-						      opts.show_all_types);
+  abixml::consider_types_not_reachable_from_public_interfaces(rdr,
+									      opts.show_all_types);
 }
 
 /// Set the regex patterns describing the functions to drop from the
@@ -929,17 +1010,17 @@ prepare_di_root_paths(options& o)
 /// @return abigail::tools_utils::ABIDIFF_ERROR if an error was
 /// detected, abigail::tools_utils::ABIDIFF_OK otherwise.
 static abigail::tools_utils::abidiff_status
-handle_error(abigail::dwarf_reader::status status_code,
-	     const abigail::dwarf_reader::read_context* ctxt,
+handle_error(abigail::fe_iface::status status_code,
+	     const abigail::elf_based_reader* rdr,
 	     const string& prog_name,
 	     const options& opts)
 {
-  if (!(status_code & abigail::dwarf_reader::STATUS_OK))
+  if (!(status_code & abigail::fe_iface::STATUS_OK))
     {
       emit_prefix(prog_name, cerr)
 	<< "failed to read input file " << opts.file1 << "\n";
 
-      if (status_code & abigail::dwarf_reader::STATUS_DEBUG_INFO_NOT_FOUND)
+      if (status_code & abigail::fe_iface::STATUS_DEBUG_INFO_NOT_FOUND)
 	{
 	  emit_prefix(prog_name, cerr) <<
 	    "could not find the debug info\n";
@@ -992,22 +1073,18 @@ handle_error(abigail::dwarf_reader::status status_code,
 	  }
 	}
 
-      if (status_code & abigail::dwarf_reader::STATUS_ALT_DEBUG_INFO_NOT_FOUND)
+      if (status_code & abigail::fe_iface::STATUS_ALT_DEBUG_INFO_NOT_FOUND)
 	{
 	  emit_prefix(prog_name, cerr)
 	    << "could not find the alternate debug info file";
-	  if (ctxt)
-	    {
-	      string alt_di_path;
-	      abigail::dwarf_reader::refers_to_alt_debug_info(*ctxt,
-							      alt_di_path);
-	      if (!alt_di_path.empty())
-		cerr << " at: " << alt_di_path;
-	    }
-	  cerr << "\n";
+
+	   if (rdr->alternate_dwarf_debug_info())
+	    cerr << " at: "
+		 << rdr->alternate_dwarf_debug_info_path()
+		 << "\n";
 	}
 
-      if (status_code & abigail::dwarf_reader::STATUS_NO_SYMBOLS_FOUND)
+      if (status_code & abigail::fe_iface::STATUS_NO_SYMBOLS_FOUND)
 	emit_prefix(prog_name, cerr)
 	  << "could not find the ELF symbols in the file '"
 	  << opts.file1
@@ -1019,6 +1096,32 @@ handle_error(abigail::dwarf_reader::status status_code,
   return abigail::tools_utils::ABIDIFF_OK;
 }
 
+/// Emit an error message saying that the two files have incompatible
+/// format versions.
+///
+/// @param file_path1 the first file path to consider.
+///
+/// @param version1 the second version to consider.
+///
+/// @param file_path2 the second file path to consider.
+///
+/// @param version2 the second version to consider.
+///
+/// @param prog_name the name of the current program.
+static void
+emit_incompatible_format_version_error_message(const string& file_path1,
+					       const string& version1,
+					       const string& file_path2,
+					       const string& version2,
+					       const string& prog_name)
+{
+  emit_prefix(prog_name, cerr)
+    << "incompatible format version between the two input files:\n"
+    << "'" << file_path1 << "' (" << version1 << ")\n"
+    << "and\n"
+    << "'" << file_path2 << "' (" << version2 << ")\n";
+}
+
 int
 main(int argc, char* argv[])
 {
@@ -1075,26 +1178,24 @@ main(int argc, char* argv[])
       abigail::tools_utils::file_type t1_type, t2_type;
 
       t1_type = guess_file_type(opts.file1);
-      if (t1_type == abigail::tools_utils::FILE_TYPE_UNKNOWN)
-	{
-	  emit_prefix(argv[0], cerr)
-	    << "Unknown content type for file " << opts.file1 << "\n";
-	  return abigail::tools_utils::ABIDIFF_ERROR;
-	}
-
       t2_type = guess_file_type(opts.file2);
-      if (t2_type == abigail::tools_utils::FILE_TYPE_UNKNOWN)
-	{
-	  emit_prefix(argv[0], cerr)
-	    << "Unknown content type for file " << opts.file2 << "\n";
-	  return abigail::tools_utils::ABIDIFF_ERROR;
-	}
 
-      environment_sptr env(new environment);
+      environment env;
+      if (opts.exported_interfaces_only.has_value())
+	env.analyze_exported_interfaces_only(*opts.exported_interfaces_only);
+
+#ifdef WITH_DEBUG_SELF_COMPARISON
+	    if (opts.do_debug_self_comparison)
+	      env.self_comparison_debug_is_on(true);
+#endif
+#ifdef WITH_DEBUG_TYPE_CANONICALIZATION
+	    if (opts.do_debug_type_canonicalization)
+	      env.debug_type_canonicalization_is_on(true);
+#endif
       translation_unit_sptr t1, t2;
-      abigail::dwarf_reader::status c1_status =
-	abigail::dwarf_reader::STATUS_OK,
-	c2_status = abigail::dwarf_reader::STATUS_OK;
+      abigail::fe_iface::status c1_status =
+	abigail::fe_iface::STATUS_OK,
+	c2_status = abigail::fe_iface::STATUS_OK;
       corpus_sptr c1, c2;
       corpus_group_sptr g1, g2;
       bool files_suppressed = false;
@@ -1119,66 +1220,62 @@ main(int argc, char* argv[])
 	  return abigail::tools_utils::ABIDIFF_ERROR;
 	  break;
 	case abigail::tools_utils::FILE_TYPE_NATIVE_BI:
-	  t1 = abigail::xml_reader::read_translation_unit_from_file(opts.file1,
-								    env.get());
+	  t1 = abixml::read_translation_unit_from_file(opts.file1,
+								       env);
 	  break;
 	case abigail::tools_utils::FILE_TYPE_ELF: // fall through
 	case abigail::tools_utils::FILE_TYPE_AR:
 	  {
-	    abigail::dwarf_reader::read_context_sptr ctxt =
-	      abigail::dwarf_reader::create_read_context
-	      (opts.file1, opts.prepared_di_root_paths1,
-	       env.get(), /*read_all_types=*/opts.show_all_types,
-	       opts.linux_kernel_mode);
-	    assert(ctxt);
+	    corpus::origin requested_fe_kind = corpus::DWARF_ORIGIN;
+#ifdef WITH_CTF
+	    if (opts.use_ctf)
+	      requested_fe_kind = corpus::CTF_ORIGIN;
+#endif
+	    abigail::elf_based_reader_sptr rdr =
+	      create_best_elf_based_reader(opts.file1,
+					   opts.prepared_di_root_paths1,
+					   env, requested_fe_kind,
+					   opts.show_all_types);
+            ABG_ASSERT(rdr);
+	    set_generic_options(*rdr, opts);
+	    set_suppressions(*rdr, opts);
+	    c1 = rdr->read_corpus(c1_status);
 
-	    abigail::dwarf_reader::set_show_stats(*ctxt, opts.show_stats);
-	    set_suppressions(*ctxt, opts);
-	    abigail::dwarf_reader::set_do_log(*ctxt, opts.do_log);
-	    c1 = abigail::dwarf_reader::read_corpus_from_elf(*ctxt, c1_status);
 	    if (!c1
 		|| (opts.fail_no_debug_info
-		    && (c1_status & STATUS_ALT_DEBUG_INFO_NOT_FOUND)
-		    && (c1_status & STATUS_DEBUG_INFO_NOT_FOUND)))
-	      return handle_error(c1_status, ctxt.get(),
+		    && (c1_status & abigail::fe_iface::STATUS_ALT_DEBUG_INFO_NOT_FOUND)
+		    && (c1_status & abigail::fe_iface::STATUS_DEBUG_INFO_NOT_FOUND)))
+	      return handle_error(c1_status, rdr.get(),
 				  argv[0], opts);
+
 	  }
 	  break;
 	case abigail::tools_utils::FILE_TYPE_XML_CORPUS:
 	  {
-	    abigail::xml_reader::read_context_sptr ctxt =
-	      abigail::xml_reader::create_native_xml_read_context(opts.file1,
-								  env.get());
-	    assert(ctxt);
-	    set_suppressions(*ctxt, opts);
-	    set_native_xml_reader_options(*ctxt, opts);
-	    c1 = abigail::xml_reader::read_corpus_from_input(*ctxt);
+	    abigail::fe_iface_sptr rdr =
+	      abixml::create_reader(opts.file1, env);
+	    assert(rdr);
+	    set_suppressions(*rdr, opts);
+	    set_native_xml_reader_options(*rdr, opts);
+	    c1 = rdr->read_corpus(c1_status);
 	    if (!c1)
-	      return handle_error(c1_status, /*ctxt=*/0,
-				  argv[0], opts);
+	      return handle_error(c1_status, /*ctxt=*/0, argv[0], opts);
 	  }
 	  break;
 	case abigail::tools_utils::FILE_TYPE_XML_CORPUS_GROUP:
 	  {
-	    abigail::xml_reader::read_context_sptr ctxt =
-	      abigail::xml_reader::create_native_xml_read_context(opts.file1,
-								  env.get());
-	    assert(ctxt);
-	    set_suppressions(*ctxt, opts);
-	    set_native_xml_reader_options(*ctxt, opts);
-	    g1 = abigail::xml_reader::read_corpus_group_from_input(*ctxt);
+	    abigail::fe_iface_sptr rdr =
+	      abixml::create_reader(opts.file1,
+							   env);
+	    assert(rdr);
+	    set_suppressions(*rdr, opts);
+	    set_native_xml_reader_options(*rdr, opts);
+	    g1 = abixml::read_corpus_group_from_input(*rdr);
 	    if (!g1)
 	      return handle_error(c1_status, /*ctxt=*/0,
 				  argv[0], opts);
 	  }
 	  break;
-	case abigail::tools_utils::FILE_TYPE_ZIP_CORPUS:
-#ifdef WITH_ZIP_ARCHIVE
-	  c1 = abigail::xml_reader::read_corpus_from_file(opts.file1);
-	  if (!c1)
-	    return handle_error(c1_status, /*ctxt=*/0, argv[0], opts);
-#endif //WITH_ZIP_ARCHIVE
-	  break;
 	case abigail::tools_utils::FILE_TYPE_RPM:
 	case abigail::tools_utils::FILE_TYPE_SRPM:
 	case abigail::tools_utils::FILE_TYPE_DEB:
@@ -1195,40 +1292,43 @@ main(int argc, char* argv[])
 	  return abigail::tools_utils::ABIDIFF_ERROR;
 	  break;
 	case abigail::tools_utils::FILE_TYPE_NATIVE_BI:
-	  t2 = abigail::xml_reader::read_translation_unit_from_file(opts.file2,
-								    env.get());
+	  t2 = abixml::read_translation_unit_from_file(opts.file2,
+								       env);
 	  break;
 	case abigail::tools_utils::FILE_TYPE_ELF: // Fall through
 	case abigail::tools_utils::FILE_TYPE_AR:
 	  {
-	    abigail::dwarf_reader::read_context_sptr ctxt =
-	      abigail::dwarf_reader::create_read_context
-	      (opts.file2, opts.prepared_di_root_paths2,
-	       env.get(), /*read_all_types=*/opts.show_all_types,
-	       opts.linux_kernel_mode);
-	    assert(ctxt);
-	    abigail::dwarf_reader::set_show_stats(*ctxt, opts.show_stats);
-	    abigail::dwarf_reader::set_do_log(*ctxt, opts.do_log);
-	    set_suppressions(*ctxt, opts);
+	    corpus::origin requested_fe_kind = corpus::DWARF_ORIGIN;
+#ifdef WITH_CTF
+	    if (opts.use_ctf)
+	      requested_fe_kind = corpus::CTF_ORIGIN;
+#endif
+            abigail::elf_based_reader_sptr rdr =
+	      create_best_elf_based_reader(opts.file2,
+					   opts.prepared_di_root_paths2,
+					   env, requested_fe_kind,
+					   opts.show_all_types);
+            ABG_ASSERT(rdr);
+
+	    set_generic_options(*rdr, opts);
+	    set_suppressions(*rdr, opts);
+
+	    c2 = rdr->read_corpus(c2_status);
 
-	    c2 = abigail::dwarf_reader::read_corpus_from_elf(*ctxt, c2_status);
 	    if (!c2
 		|| (opts.fail_no_debug_info
-		    && (c2_status & STATUS_ALT_DEBUG_INFO_NOT_FOUND)
-		    && (c2_status & STATUS_DEBUG_INFO_NOT_FOUND)))
-	      return handle_error(c2_status, ctxt.get(), argv[0], opts);
-
+		    && (c2_status & abigail::fe_iface::STATUS_ALT_DEBUG_INFO_NOT_FOUND)
+		    && (c2_status & abigail::fe_iface::STATUS_DEBUG_INFO_NOT_FOUND)))
+	      return handle_error(c2_status, rdr.get(), argv[0], opts);
 	  }
 	  break;
 	case abigail::tools_utils::FILE_TYPE_XML_CORPUS:
 	  {
-	    abigail::xml_reader::read_context_sptr ctxt =
-	      abigail::xml_reader::create_native_xml_read_context(opts.file2,
-								  env.get());
-	    assert(ctxt);
-	    set_suppressions(*ctxt, opts);
-	    set_native_xml_reader_options(*ctxt, opts);
-	    c2 = abigail::xml_reader::read_corpus_from_input(*ctxt);
+	    abigail::fe_iface_sptr rdr = abixml::create_reader(opts.file2, env);
+	    assert(rdr);
+	    set_suppressions(*rdr, opts);
+	    set_native_xml_reader_options(*rdr, opts);
+	    c2 = rdr->read_corpus(c2_status);
 	    if (!c2)
 	      return handle_error(c2_status, /*ctxt=*/0, argv[0], opts);
 
@@ -1236,24 +1336,15 @@ main(int argc, char* argv[])
 	  break;
 	case abigail::tools_utils::FILE_TYPE_XML_CORPUS_GROUP:
 	  {
-	    abigail::xml_reader::read_context_sptr ctxt =
-	      abigail::xml_reader::create_native_xml_read_context(opts.file2,
-								  env.get());
-	    assert(ctxt);
-	    set_suppressions(*ctxt, opts);
-	    set_native_xml_reader_options(*ctxt, opts);
-	    g2 = abigail::xml_reader::read_corpus_group_from_input(*ctxt);
+	    abigail::fe_iface_sptr rdr = abixml::create_reader(opts.file2, env);
+	    assert(rdr);
+	    set_suppressions(*rdr, opts);
+	    set_native_xml_reader_options(*rdr, opts);
+	    g2 = abixml::read_corpus_group_from_input(*rdr);
 	    if (!g2)
 	      return handle_error(c2_status, /*ctxt=*/0, argv[0], opts);
 	  }
 	  break;
-	case abigail::tools_utils::FILE_TYPE_ZIP_CORPUS:
-#ifdef WITH_ZIP_ARCHIVE
-	  c2 = abigail::xml_reader::read_corpus_from_file(opts.file2);
-	  if (!c2)
-	    return handle_error(c2_status, /*ctxt=*/0, argv[0], opts);
-#endif //WITH_ZIP_ARCHIVE
-	  break;
 	case abigail::tools_utils::FILE_TYPE_RPM:
 	case abigail::tools_utils::FILE_TYPE_SRPM:
 	case abigail::tools_utils::FILE_TYPE_DEB:
@@ -1300,6 +1391,18 @@ main(int argc, char* argv[])
 	      return abigail::tools_utils::ABIDIFF_OK;
 	    }
 
+	  const auto c1_version = c1->get_format_major_version_number();
+	  const auto c2_version = c2->get_format_major_version_number();
+	  if (c1_version != c2_version)
+	    {
+	      emit_incompatible_format_version_error_message(opts.file1,
+							     c1_version,
+							     opts.file2,
+							     c2_version,
+							     argv[0]);
+	      return abigail::tools_utils::ABIDIFF_ERROR;
+	    }
+
 	  set_corpus_keep_drop_regex_patterns(opts, c1);
 	  set_corpus_keep_drop_regex_patterns(opts, c2);
 
@@ -1322,6 +1425,18 @@ main(int argc, char* argv[])
 	      return abigail::tools_utils::ABIDIFF_OK;
 	    }
 
+	  const auto g1_version = g1->get_format_major_version_number();
+	  const auto g2_version = g2->get_format_major_version_number();
+	  if (g1_version != g2_version)
+	    {
+	      emit_incompatible_format_version_error_message(opts.file1,
+							     g1_version,
+							     opts.file2,
+							     g2_version,
+							     argv[0]);
+	      return abigail::tools_utils::ABIDIFF_ERROR;
+	    }
+
 	  adjust_diff_context_for_kmidiff(*ctxt);
 	  corpus_diff_sptr diff = compute_diff(g1, g2, ctxt);
 
